/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Type;
import java.util.Collection;

import org.apache.commons.beanutils.ConvertUtils;
import org.apache.commons.lang.ClassUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code AbstractParameter} is an abstract base class of {@code Parameter}.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public abstract class AbstractParameter implements Parameter {

	/** Logger for this class. */
	protected final Logger logger = LoggerFactory.getLogger(getClass());
	
	/** Web context. */
	protected final WebContext context;
	
	/** The parameter type. */
	protected final Type type;
	
	/** The parameter name. */
	protected final String name;
	
	/** The parameter value */
	protected Object value;
	
	/**
	 * Constructs a new {@code AbstractParameter} from the specified Web 
	 * context, parameter type and parameter name.
	 * If this constructor is invoked, the parameter value is constructed 
	 * immediately. 
	 * The parameter value construction is described at {@code #construct()} 
	 * method. Check {@link #construct()} method carefully. Built-in parameter 
	 * implementations such as {@code QueryParameter}, {@code BodyParameter}, 
	 * {@code HeaderParameter}, {@code CookieParameter}, 
	 * {@code SessionParameter} and {@code ApplicationParameter} extend this 
	 * class and provide the same construction mechanism.
	 * 
	 * @param context The Web context.
	 * @param type The parameter type.
	 * @param name The parameter name.
	 * @see AbstractParameter#construct()
	 */
	protected AbstractParameter(WebContext context, Type type, String name) {
		this(context, type, name, true);
	}
	
	/**
	 * Constructs a new {@code AbstractParameter} from the specified Web 
	 * context, parameter type and parameter name.
	 * If this constructor is invoked, the timing of parameter value 
	 * construction is able to be controlled by each subclass (with 
	 * <code>constructs</code> parameter).
	 * The parameter value construction is described at {@code #construct()} 
	 * method. Check {@link #construct()} method carefully. Built-in parameter 
	 * implementations such as {@code QueryParameter}, {@code BodyParameter}, 
	 * {@code HeaderParameter}, {@code CookieParameter}, 
	 * {@code SessionParameter} and {@code ApplicationParameter} extend this 
	 * class and provide the same construction mechanism.
	 * 
	 * @param context The Web context.
	 * @param type The parameter type.
	 * @param name The parameter name.
	 * @param constructs Whether the parameter value is constructed immediately 
	 * or not. <code>true</code> indicates 'Constructed immediately'.
	 * @see AbstractParameter#construct()
	 */
	protected AbstractParameter(WebContext context, Type type, String name, boolean constructs) {
		this.context = context;
		this.type = type;
		this.name = name;
		
		if (constructs) {
			construct();
		}
	}
	
	/**
	 * Returns the value of Web endpoint parameter.
	 * 
	 * @return The value of Web endpoint parameter.
	 */
	@Override
	public Object getValue() {
		return value;
	}
	
	/**
	 * Constructs the value of Web endpoint parameter.
	 * Web endpoint parameter declaration falls into the following patterns and 
	 * each of the patterns has the own construction logic: 
	 * <ol>
	 * <li>Named array type - Unsupported. Always returns <code>null</code>.</li>
	 * <li>Named collection of core value type - Supported. The collection 
	 * instance is constructed according to the default implementation type 
	 * (See {@code Types#getDefaultImplementationType(Type)}). If the collection 
	 * type is not supported (See {@code Types#isSupportedCollection(Type)}), 
	 * this method returns <code>null</code>. Each of the collection elements is 
	 * constructed from the values that {@code #getValues(String)} has returned 
	 * as the same way of the following core value type. If the 
	 * constructed collection has no element, this method returns <code>null
	 * </code>.</li>
	 * <li>Named collection of user-defined value type - Supported. The 
	 * collection instance is constructed according to the default 
	 * implementation type (as the same as collection of core value type). Each 
	 * of the collection elements is constructed from the values that 
	 * {@code #getValues(String)} has returned as the same way of following 
	 * user-defined value type.
	 * </li>
	 * <li>Named collection of user-defined object type - Unsupported. Always 
	 * returns <code>null</code>.</li>
	 * <li>Named core value type - Supported. The value that {@code #getValue()} 
	 * has returned is converted to the core value type. If the conversion 
	 * failed, returns <code>null</code></li>
	 * <li>Named user-defined value type - Supported. If the value that 
	 * {@code #getValue()} has returned is assignable to the user-defined value 
	 * type, returns it. If the value that {@code #getValue()} has returned is 
	 * {@code String.class}, this method constructs the user-defined value type 
	 * with public constructor that takes one String.class parameter or public 
	 * static <code>valueOf(String.class)</code> method. If the conversion 
	 * failed, returns <code>null</code>.
	 * </li>
	 * <li>Named user-defined object type - Partially supported. If the value 
	 * that {@code #getValue()} method has returned is assignable to the 
	 * user-defined object type, returns it. Otherwise, retunrs <code>null</code> 
	 * (does not any type conversion). </li>
	 * <li>Not named array type - Unsupported. Always returns <code>null</code>.
	 * </li>
	 * <li>Not named collection of core value type - Unsupported. Always returns 
	 * <code>null</code>.</li>
	 * <li>Not named collection of user-defined value type - Unsupported. Always 
	 * returns <code>null</code>.</li>
	 * <li>Not named collection of user-defined object type - Unsupported. 
	 * Always returns <code>null</code>.</li>
	 * <li>Not named core value type - Unsupported. Always returns 
	 * <code>null</code>.</li>
	 * <li>Not named user-defined value type - Unsupported. Always returns 
	 * <code>null</code>.</li>
	 * <li>Not named user-defined object type - Supported. First, this method 
	 * instantiates the user-defined object instance form public default 
	 * constructor, and then constructs each of the instances' fields value 
	 * according to the named type construction described above (by invoking 
	 * {@code #newParameter(WebContext, Type, String)} with the field 
	 * type and field name). If the instantiation is failed, returns <code>null
	 * </code>.
	 * </li>
	 * </ol>
	 * This method is overridable. You can provide your own value construction 
	 * to this parameter by overriding this method.
	 */
	@SuppressWarnings("unchecked")
	protected void construct() {
		if (name != null && !name.isEmpty()) {
			// Named parameter construction.
			if (Types.isArray(type)) {
				// Named array type.
				logger.debug("Array type [" + type + "] is not supported in ["
						+ getClass() + "]");
			} else if (Types.isCollection(type)) {
				// Named collection type.
				if (Types.isSupportedCollection(type)) {
					Collection<Object> values = getValues(name);
					
					if (values == null) {
						logger.debug("Collection named [" + name + "] not found");
					} else {
						Class<?> elementType = Types.getElementType(type);
						boolean coreValueType = Types.isCoreValueType(elementType);
						
						if (!coreValueType && !Types.isUserDefinedValueType(elementType)) {
							// Named collection of user-defined object.
							logger.debug("Collection element type ["
									+ elementType + "] is not supported in ["
									+ getClass() + "]");
						} else {
							try {
								Class<?> implementationType = Types.getDefaultImplementationType(type);
								Collection<Object> collection = (Collection<Object>) implementationType.newInstance();
								
								for (Object object : values) {
									// Named collection of core value type.
									// Named collection of user-defined value type.
									Object convert = (coreValueType) ? convert(object, elementType)
											: convertUserDefinedValueType(object, elementType);
									
									if (convert != null
											&& ClassUtils.primitiveToWrapper(elementType)
													.isAssignableFrom(convert.getClass())) {
										collection.add(convert);
									} else {
										logger.debug("Parameter [" + convert
												+ "] cannot be converted to ["
												+ elementType + "]");
									}
								}
								
								value = (!collection.isEmpty()) ? collection : null;
							} catch (Exception e) {
								logger.debug("Cannot instantiate ["
										+ Types.getDefaultImplementationType(type)
										+ "] (Default implementation type of ["
										+ type + "])", e);
							}
						}
					}
					
				} else {
					logger.debug("Parameter type [" + type
							+ "] is not supported in [" + getClass() + "]");
				}
				
			} else if (Types.isCoreValueType(type)) {
				// Named core value type.
				Class<?> rawType = Types.getRawType(type);
				Object object = getValue(name);
				
				if (object == null) {
					logger.debug("Value named [" + name + "] not found");
				} else {
					Object convert = convert(object, rawType);
					
					if (convert != null
							&& ClassUtils.primitiveToWrapper(rawType)
									.isAssignableFrom(convert.getClass())) {
						value = convert;
					} else {
						logger.warn("Parameter [" + convert
								+ "] cannot be converted to [" + type + "]");
					}
				}
				
			} else if (Types.isUserDefinedValueType(type)) {
				// Named user-defined value type.
				Object object = getValue(name);
				Class<?> rawType = Types.getRawType(type);
				
				if (object == null) {
					logger.debug("Value named [" + name + "] not found");
				} else {
					Object userDefinedValueType = convertUserDefinedValueType(object, rawType);
					
					if (userDefinedValueType == null) {
						logger.warn("Parameter [" + object
								+ "] cannot be converted to [" + type + "]");
					}
					
					value = userDefinedValueType;
				}
				
			} else {
				// Named user-defined object type.
				Object object = getValue(name);
				
				if (object == null) {
					logger.debug("Value named [" + name + "] not found");
				} else if (Types.getRawType(type).isAssignableFrom(object.getClass())) {
					value = object;
				} else {
					logger.warn("Parameter [" + object
							+ "] cannot be converted to [" + type + "]");
				}
			}
			
		} else {
			// Non-named parameter construction.
			if (Types.isArray(type) || Types.isCollection(type)
					|| Types.isCoreValueType(type)
					|| Types.isUserDefinedValueType(type)) {
				// Not named array type.
				// Not named collection (of core value type or user-defined object type).
				// Not named core value type.
				// Not named user-defined value type.
				logger.debug("Non-named parameter type [" + type
						+ "] is not supported in [" + getClass() + "]");
			} else {
				// Not named user-defined object type.
				Class<?> rawType = Types.getRawType(type);
				
				try {
					Object instance = rawType.newInstance();
					
					for (Field field : rawType.getDeclaredFields()) {
						Object object = newParameter(context, 
								field.getGenericType(), field.getName()).getValue();
						
						if (object != null) {
							field.setAccessible(true);
							field.set(instance, object);
						}
					}
					
					value = instance;
				} catch (Exception e) {
					logger.warn("Cannot instantiate [" + type + "]", e);
				}
			}
		}
	}
	
	/**
	 * Converts the specified object to the specified type.
	 * This method is overridable. You can provide your own conversion to the 
	 * parameter construction by overriding this method.
	 * 
	 * @param object The object to be converted
	 * @param type The type to which the specified object be converted.
	 * @return The converted object.
	 */
	protected Object convert(Object object, Class<?> type) {
		try {
			return ConvertUtils.convert(object, type);
		} catch (Exception e) {
			logger.warn("Cannot convert [" + object + "] to [" + type + "]", e);
			return null;
		}
	}
	
	/**
	 * Converts the specified object to the specified user-defined value type.
	 * This method is overridable. You can provide your own conversion to the 
	 * parameter construction by overriding this method.
	 * 
	 * @param object The object to be converted.
	 * @param type The type to which the specified object be converted.
	 * @return The converted object.
	 */
	protected Object convertUserDefinedValueType(Object object, Class<?> type) {
		if (type.isAssignableFrom(object.getClass())) {
			return object;
		} else if (object instanceof String) {
			try {
				Constructor<?> constructor = type.getConstructor(String.class);
				return constructor.newInstance(object);
			} catch (Exception e) {
				logger.debug("Cannot invoke [public " + type.getName() 
						+ "(String.class)] constrcutor on [" + type + "]", e);
			}
			
			try {
				return type.getMethod("valueOf", String.class).invoke(null, object);
			} catch (Exception e1) {
				logger.debug("Cannot invoke [public static " 
						+ type.getName() + ".valueOf(String.class)]" 
						+ "method on [" + type + "]", e1);
			}
			
		} else {
			logger.warn("Parameter [" + object
					+ "] cannot be converted to [" + type + "]");
		}
		
		return null;
	}
	
	/**
	 * Constructs a new {@code Parameter} instance with the specified Web 
	 * context, parameter type and parameter name.
	 * 
	 * @param context Web context.
	 * @param type The parameter type.
	 * @param name The parameter name.
	 * @return A {@code Parameter} instance.
	 */
	protected abstract Parameter newParameter(WebContext context, Type type, String name);
	
	/**
	 * Returns the value corresponding to the specified name.
	 * The returned value is passed to {@code #convert(Object, Class)} method 
	 * and converted to appropriate parameter type.
	 * 
	 * @param name The name corresponding to the value.
	 * @return The value corresponging to the specified name.
	 */
	protected abstract Object getValue(String name);
	
	/**
	 * Returns the value corresponding to the specified name as 
	 * {@code Collection} view. 
	 * Each of the returned collection's values is passed to 
	 * {@code #convert(Object, Class)} method and converted to appropriate 
	 * element type of the collection parameter type.
	 * 
	 * @param name The name corresponding to the value.
	 * @return The value corresponding to the specified name as 
	 * {@code Collection} view.
	 */
	protected abstract Collection<Object> getValues(String name);
	
}