/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony;

import org.eiichiro.monophony.json.JSONRequest;
import org.eiichiro.monophony.json.JSONResponse;
import org.eiichiro.monophony.xml.XMLRequest;
import org.eiichiro.monophony.xml.XMLResponse;
import org.eiichiro.reverb.lang.UncheckedException;

/**
 * {@code DefaultSettings} provides Monophony framework's default settings.
 * If you would not specify your own settings in web.xml, Monophony uses this 
 * settings by default.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class DefaultSettings implements Settings {

	private Loader loader = new WebInfClassesLoader();
	
	private Instantiator instantiator = new Instantiator() {
		
		@Override
		public <T> T instantiate(Class<T> endpoint) {
			try {
				return endpoint.newInstance();
			} catch (Exception e) {
				throw new UncheckedException(e);
			}
		}
		
	};
	
	private static DefaultSettings settings = new DefaultSettings();
	
	/**
	 * Returns singleton {@code DefaultSettings} instance.
	 * @return The singleton {@code DefaultSettings} instance.
	 */
	public static DefaultSettings newInstance() {
		return settings;
	}
	
	/**
	 * Returns cached {@code WebInfClassesLoader} instance.
	 * 
	 * @see org.eiichiro.monophony.Settings#loader()
	 */
	@Override
	public Loader loader() {
		return loader;
	}

	/**
	 * Returns the default request type corresponding to the specified MIME 
	 * media type.
	 * <ol>
	 * <li>{@code application/xml} -> {@code XMLRequest.class}</li>
	 * <li>{@code application/json} -> {@code JSONRequest.class}</li>
	 * <li>Other -> {@code GenericRequest.class}</li>
	 * </ol>
	 * 
	 * @see org.eiichiro.monophony.Settings#requestType(java.lang.String)
	 */
	@Override
	public Class<? extends Request> requestType(String contentType) {
		if (contentType == null) {
			return GenericRequest.class;
		}
		
		if (contentType.startsWith(MediaType.APPLICATION_JSON)) {
			return JSONRequest.class;
		} else if (contentType.startsWith(MediaType.APPLICATION_XML)) {
			return XMLRequest.class;
		} else {
			// XXX: Qualifying more.
			return GenericRequest.class;
		}
	}

	/**
	 * Returns the default response type corresponding to the specified MIME 
	 * media type.
	 * <ol>
	 * <li>{@code application/xml} -> {@code XMLResponse.class}</li>
	 * <li>{@code application/json} -> {@code JSONResponse.class}</li>
	 * <li>Other -> {@code GenericResponse.class}</li>
	 * </ol>
	 * 
	 * @see org.eiichiro.monophony.Settings#responseType(java.lang.String)
	 */
	@Override
	public Class<? extends Response> responseType(String contentType) {
		if (contentType == null) {
			return GenericResponse.class;
		}
		
		if (contentType.startsWith(MediaType.APPLICATION_JSON)) {
			return JSONResponse.class;
		} else if (contentType.startsWith(MediaType.APPLICATION_XML)) {
			return XMLResponse.class;
		} else {
			// XXX: Qualifying more.
			return GenericResponse.class;
		}
	}

	@Override
	public Instantiator instantiator() {
		return instantiator;
	}

}
