/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import javax.servlet.http.HttpServletResponse;

import org.eiichiro.reverb.lang.UncheckedException;

/**
 * {@code GenericResponse} is a generic implementation of {@code Response}.
 * If the response type corresponding to the MIME media type specified on 
 * Web endpoint method is not specified by {@code Settings}, Monophony uses 
 * this implementation. UTF-8 is used for the character encoding (MIME charset) 
 * at any time.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class GenericResponse implements Response {

	/** Web endpoint invocation result. */
	protected final Result result;
	
	/** MIME media type. */
	protected String mediaType = "";
	
	/** HTTP status code. */
	protected int status = 200;
	
	/**
	 * Constructs a new {@code GenericResponse} with the specified Web endpoint 
	 * invocation result.
	 * 
	 * @param result Web endpoint invocation result from which this 
	 * {@code Response} is built.
	 */
	public GenericResponse(Result result) {
		this.result = result;
	}
	
	/**
	 * Writes HTTP response representation to output stream.
	 * This method determines the response form as the following convention: 
	 * <ul>
	 * <li>
	 * If the MIME media type ({@code #mediaType}) is not empty, set it to 
	 * 'Content-Type' HTTP response header.
	 * </li>
	 * <li>
	 * If the status ({@code #status}) is greater than 0, set it to HTTP 
	 * response code.
	 * </li>
	 * <li>
	 * If {@code #value} is instance of {@code String}, this method writes it to 
	 * the HTTP response directly. If MIME media type has not been specified, 
	 * 'text/plain' is used.
	 * </li>
	 * <li>
	 * If {@code #value} is instance of {@code Externalizable}, this method 
	 * writes it to the HTTP response as {@code ObjectOutputStream} by invoking 
	 * {@code Externalizable#writeExternal(java.io.ObjectOutput)}.
	 * </li>
	 * <li>
	 * If {@code #value} is instance of {@code Serializable}, this method writes 
	 * it to the HTTP response as {@code ObjectOutputStream}.
	 * </li>
	 * <li>
	 * The other, this method write it to the HTTP response as plain text by 
	 * invoking {@code #value#toString()}. If MIME media type has not been 
	 * specified, 'text/plain' is used.
	 * </li>
	 * </ul>
	 * 
	 * @param context Web context.
	 */
	@Override
	public void out(WebContext context) {
		HttpServletResponse response = context.getResponse();
		
		if (!mediaType.isEmpty()) {
			response.setHeader("Content-Type", mediaType);
		}
		
		if (status > 0) {
			response.setStatus(status);
		}
		
		Object value = result.getValue();
		
		if (value == null) {
			return;
		}
		
		try {
			if (value instanceof String) {
				if (mediaType.isEmpty()) {
					response.setHeader("Content-Type", MediaType.TEXT_PLAIN);
				}
				
				response.setCharacterEncoding("UTF-8");
				response.getWriter().write((String) value);
			} else if (value instanceof Serializable) {
				new ObjectOutputStream(response.getOutputStream()).writeObject(value);
			} else {
				if (mediaType.isEmpty()) {
					response.setHeader("Content-Type", MediaType.TEXT_PLAIN);
				}
				
				response.setCharacterEncoding("UTF-8");
				response.getWriter().write(value.toString());
			}
			
		} catch (IOException e) {
			throw new UncheckedException(e);
		}
	}
	
	@Override
	public Response mediaType(String mediaType) {
		if (mediaType == null) {
			throw new IllegalArgumentException("'mediaType' must not be [" + mediaType + "]");
		}
		
		this.mediaType = mediaType;
		return this;
	}

	@Override
	public Response status(int status) {
		this.status = status;
		return this;
	}

}
