/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony;

import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.eiichiro.reverb.reflection.MethodInvocation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code MonophonyFilter} is a central Servlet {@code Filter} to process Web 
 * accesses by Monophony.
 * You need to setup this filter in web.xml in order to use Monophony, as the 
 * following: 
 * <pre>
 * &lt;filter&gt;
 * 	&lt;filter-name&gt;monophony&lt;/filter-name&gt;
 * 	&lt;filter-class&gt;org.eiichiro.monophony.MonophonyFilter&lt;/filter-class&gt;
 * &lt;/filter&gt;
 * &lt;filter-mapping&gt;
 * 	&lt;filter-name&gt;monophony&lt;/filter-name&gt;
 * 	&lt;url-pattern&gt;/*&lt;/url-pattern&gt;
 * &lt;/filter-mapping&gt;
 * </pre>
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class MonophonyFilter implements Filter {

	/** Servlet context parameter key to get custom {@code Settings} class. */
	public static final String SETTINGS = "org.eiichiro.monophony.settings";
	
	private static Logger logger = LoggerFactory.getLogger(MonophonyFilter.class);
	
	private Settings settings;
	
	private Receiver receiver;
	
	private Sender sender;
	
	static {
		logger.info("Monophony " + Version.MAJOR + "." + Version.MINER + "." + Version.BUILD);
		logger.info("Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.");
	}
	
	/** Destroys this fileter. */
	@Override
	public void destroy() {}

	/**
	 * Processes the Web request with Monophony.
	 * 
	 * @param request HTTP request.
	 * @param response HTTP response.
	 * @throws IOException If any I/O error has occurred.
	 * @throws ServletException If any exception has occurred in processing the 
	 * request.
	 */
	@Override
	public void doFilter(ServletRequest request, ServletResponse response,
			FilterChain chain) throws IOException, ServletException {
		HttpServletRequest req = (HttpServletRequest) request;
		HttpServletResponse res = (HttpServletResponse) response;
		String uri = req.getRequestURI();
		
		if (uri.substring(uri.lastIndexOf("/")).contains(".")) {
			chain.doFilter(request, response);
		} else {
			WebContext context = new WebContext(req, res);
			
			try {
				MethodInvocation<?> invocation = receiver.receive(context);
				
				if (invocation == null) {
					// If Web endpoint class is not found, just chains the filter.
					chain.doFilter(request, response);
				} else {
					sender.send(context, invocation);
				}
				
			} catch (CannotAcceptRequestException e) {
				logger.warn("Failed to accept HTTP request", e);
				res.sendError(e.getStatus());
				return;
			} catch (CannotSendResponseException e) {
				logger.warn("Failed to send HTTP response", e);
				res.sendError(e.getStatus());
				return;
			}
		}
	}

	/**
	 * Initializes this filter with the specified configuration.
	 * This methods attempts to get custom settings class from {@code ServletContext} 
	 * init parameter and instantiate it first. If the custom settings class has 
	 * been specified, Monophony is setup according to the custom settings.
	 * Next, starts request dispatcher specifying the custom settings (if specified) 
	 * and default settings.
	 * 
	 * @param filterConfig Servlet filter configuration.
	 */
	@Override
	public void init(FilterConfig filterConfig) throws ServletException {
		logger.info("Initializing...");
		settings = getSettings(filterConfig);
		receiver = new Receiver(settings);
		logger.info("Request receiver started");
		sender = new Sender(settings);
		logger.info("Response sender started");
		logger.info("Initializing completed");
	}
	
	protected Settings getSettings(FilterConfig filterConfig) throws ServletException {
		ServletContext context = filterConfig.getServletContext();
		String clazz = context.getInitParameter(SETTINGS);
		
		if (clazz != null) {
			try {
				Settings settings = (Settings) Class.forName(clazz).newInstance();
				logger.info("Custom settings [" + clazz + "] loaded");
				return settings;
			} catch (Exception e) {
				logger.error("Failed to load custom settings ["
						+ clazz
						+ "]; Does the class have a public default constructor?");
				throw new ServletException(e);
			}
		}
		
		return new Settings() {
			
			@Override
			public Class<? extends Response> responseType(String contentType) {
				return null;
			}
			
			@Override
			public Class<? extends Request> requestType(String contentType) {
				return null;
			}
			
			@Override
			public Loader loader() {
				return null;
			}

			@Override
			public Instantiator instantiator() {
				return null;
			}
			
		};
	}

}
