/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.eiichiro.monophony.annotation.Accepts;
import org.eiichiro.monophony.annotation.Default;
import org.eiichiro.monophony.annotation.Endpoint;
import org.eiichiro.monophony.annotation.GET;
import org.eiichiro.monophony.annotation.POST;
import org.eiichiro.monophony.annotation.Source;
import org.eiichiro.reverb.lang.UncheckedException;
import org.eiichiro.reverb.reflection.MethodInvocation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Receiver} is a core component to validate HTTP request and build Web 
 * endpoint method invocation as {@code Invocation}.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Receiver {

	/** Logger for {@code Receiver}. */
	protected Logger logger = LoggerFactory.getLogger(getClass());
	
	/** Monophony settings. */
	protected final Settings settings;
	
	/** URL and Web endpoint class map. */
	protected Map<String, Class<?>> endpoints = new HashMap<String, Class<?>>();

	/**
	 * Constructs a new {@code Receiver} with the specified Monophony settings.
	 * 
	 * @param settings Monophony settings.
	 */
	public Receiver(Settings settings) {
		this.settings = settings;
		Loader loader = settings.loader();
		
		if (loader == null) {
			loader = DefaultSettings.newInstance().loader();
		}
		
		for (Class<?> clazz : loader.load()) {
			Endpoint endpoint = clazz.getAnnotation(Endpoint.class);
			
			if (endpoint == null) {
				logger.warn("Class [" + clazz
						+ "] is not annotated with @Endpoint; "
						+ "Cannot detect as Web endpoint");
				continue;
			}
			
			String uri = endpoint.value();
			
			if (uri.isEmpty()) {
				uri = "/" + clazz.getSimpleName().toLowerCase();
			}
			
			Class<?> duplicated = endpoints.get(uri);
			
			if (duplicated != null) {
				logger.warn("Web endpoint corresponding to [" + uri
						+ "] is duplicated: [" + duplicated
						+ "] is overwritten by [" + clazz + "]");
			}
			
			endpoints.put(uri, clazz);
		}
		
		if (logger.isDebugEnabled()) {
			for (String uri : endpoints.keySet()) {
				logger.debug("Web endpoint [" + endpoints.get(uri)
						+ "] handles [" + uri + "]");
			}
		}
		
		logger.info("Web endpoints loaded");
	}
	
	/**
	 * Validates HTTP request and build Web endpoint method invocation.
	 * This method processes the request as the following steps: 
	 * <ol>
	 * <li>Identifies Web endpoint class from requested path. For example, the 
	 * requested path is <code>'http://example.com/app/path/to/endpoint/doit'</code>, 
	 * this class gets Web endpoint class from 'path/to/endpoint'. If the Web 
	 * endpoint class is not found, returns <code>null</code>.
	 * </li>
	 * <li>Identifies Web endpoint method from requested path. If the requested 
	 * path is the same as above, gets the method named 'doit' from the Web 
	 * endpoint class. If the requested path ends with '/', this class assumes 
	 * {@code @Default} annotated method as the Web endpoint method to 
	 * invoke.</li>
	 * <li>Checks the requested HTTP method is acceptable. If the Web endpoint 
	 * method is qualified by {@code @GET} or {@code @POST}, this class accepts 
	 * only the qualified HTTP method.</li>
	 * <li>Checks the requested MIME media type is acceptable. If the Web 
	 * endpoint method is qualified by {@code @Accepts}, this class accepts only 
	 * the qualified MIME media types.</li>
	 * <li>Determines {@code Request} type from the content type on HTTP request 
	 * header by invoking {@code Settings#requestType(String)} method with the 
	 * content type.</li>
	 * <li>Constructs Web endpoint method parameters by invoking {@code Request#get()}
	 * methods.</li>
	 * <li>Constructs Web endpoint {@code MethodInvocation} from the class, 
	 * method and the parameters.
	 * </li>
	 * </ol>
	 * 
	 * @param context Web context.
	 * @return Web endpoint method invocation.
	 * @throws CannotAcceptRequestException If the HTTP request cannot be 
	 * accepted for any reason.
	 */
	public MethodInvocation<?> receive(WebContext context)
			throws CannotAcceptRequestException {
		HttpServletRequest req = context.getRequest();
		String path = req.getServletPath();
		int index = path.lastIndexOf("/");
		String endpoint = path.substring(0, index);
		Class<?> clazz = endpoints.get(endpoint);
		
		if (logger.isDebugEnabled()) {
			logger.debug("Request path is [" + path + "]");
			logger.debug("Web endpoint is [" + clazz + "]");
		}

		if (clazz == null) {
			if (logger.isDebugEnabled()) {
				logger.debug("Web endpoint class not found: Request path is [" + path + "]");
			}
			
			return null;
		}
		
		String methodName = (path.endsWith("/")) ? "" : path.substring(index + 1);
		
		for (Method method : clazz.getMethods()) {
			// Method mapping.
			if (methodName.isEmpty()) {
				if (method.getAnnotation(Default.class) == null) {
					continue;
				}
				
			} else if (method.getName().compareToIgnoreCase(methodName) != 0) {
				continue;
			}
			
			logger.trace("Endpoint method is ["
					+ method.getDeclaringClass().getName() + "#"
					+ method.getName() + "]");
			
			// Is the HTTP method acceptable?
			boolean get = (method.getAnnotation(GET.class) != null);
			boolean post = (method.getAnnotation(POST.class) != null);
			String httpMethod = req.getMethod();
			
			if (get && post) {
				if (!httpMethod.equals("GET") && !httpMethod.equals("POST")) {
					throw new CannotAcceptRequestException(
							HttpServletResponse.SC_METHOD_NOT_ALLOWED,
							"Endpoint method ["
									+ method.getDeclaringClass().getName() + "#" + method.getName()
									+ "] is capable of accepting [GET] or [POST] method");
				}
				
			} else if (get) {
				if (!httpMethod.equals("GET")) {
					throw new CannotAcceptRequestException(
							HttpServletResponse.SC_METHOD_NOT_ALLOWED,
							"Endpoint method ["
									+ method.getDeclaringClass().getName() + "#" + method.getName()
									+ "] is capable of accepting only [GET] method");
				}
				
			} else if (post) {
				if (!httpMethod.equals("POST")) {
					throw new CannotAcceptRequestException(
							HttpServletResponse.SC_METHOD_NOT_ALLOWED,
							"Endpoint method ["
									+ method.getDeclaringClass().getName() + "#" + method.getName()
									+ "] is capable of accepting only [POST] method");
				}
			}
			
			// Is the MIME media type of the request acceptable?
			Accepts accepts = method.getAnnotation(Accepts.class);
			String type = req.getContentType();
			
			if (accepts != null) {
				boolean acceptable = false;
				
				for (String value : accepts.value()) {
					if (type.startsWith(value)) {
						acceptable = true;
						break;
					}
				}
				
				if (!acceptable) {
					throw new CannotAcceptRequestException(
							HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE,
							"Endpoint method [" + method.getDeclaringClass().getName() 
									+ "#" + method.getName()
									+ "] is not capable of accepting [" + type
									+ "] media type");
				}
			}
			
			// Request type determination.
			Request request = null;
			Class<? extends Request> requestType = settings.requestType(type);
			
			if (requestType == null) {
				requestType = DefaultSettings.newInstance().requestType(type);
			}
			
			try {
				Constructor<? extends Request> constructor 
						= requestType.getConstructor(WebContext.class);
				request = constructor.newInstance(context);
			} catch (Exception e) {
				throw new UncheckedException(e);
			}
			
			// Parameter construction.
			List<Object> parameters = new ArrayList<Object>();
			Annotation[][] parameterAnnotations = method.getParameterAnnotations();
			Type[] parameterTypes = method.getGenericParameterTypes();
			
			for (int i = 0; i < parameterTypes.length; i++) {
				List<Annotation> sources = new ArrayList<Annotation>();
				
				for (Annotation annotation : parameterAnnotations[i]) {
					if (annotation.annotationType().getAnnotation(Source.class) != null) {
						sources.add(annotation);
					}
				}
				
				parameters.add(request.get(parameterTypes[i], sources).getValue());
			}
			
			Instantiator instantiator = settings.instantiator();
			
			if (instantiator == null) {
				instantiator = DefaultSettings.newInstance().instantiator();
			}
			
			return new MethodInvocation<Object>(method, 
					instantiator.instantiate(clazz), parameters.toArray());
		}
		
		throw new CannotAcceptRequestException(
				HttpServletResponse.SC_NOT_FOUND,
				"Web endpoint method not found: Request path is [" + path + "]");
	}
	
}
