/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony;

import org.eiichiro.monophony.json.JSONResponse;
import org.eiichiro.monophony.xml.XMLResponse;

/**
 * {@code Send} is the utility which has some factory methods to create 
 * {@code Response} instance to be sent to the client form Web endpoint.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public abstract class Send {

	private Send() {}

	/**
	 * Creates generic {@code Response} that has no message body.
	 * 
	 * @return Generic {@code Response} that has no message body.
	 */
	public static Response response() {
		return new GenericResponse(new Result(null));
	}

	/**
	 * Creates generic {@code Response} that has the specified entity as message 
	 * body.
	 * 
	 * @param value The object to be sent to the client.
	 * @return Generic {@code Response} that has the specified entity as message 
	 * body.
	 */
	public static Response response(Object value) {
		return new GenericResponse(new Result(value));
	}
	
	/**
	 * Creates forward {@code Response} to forward the request to the specified 
	 * path.
	 * 
	 * @param path The path to forward.
	 * @return {@code Forward} to forward the request to the specified path.
	 */
	public static Forward forward(String path) {
		return new Forward(path);
	}
	
	/**
	 * Creates redirect {@code Response} to redirect the response to the 
	 * specified URL.
	 * 
	 * @param url The URL to redirect.
	 * @return {@code Redirect} to redirect the response to the specified URL.
	 */
	public static Redirect redirect(String url) {
		return new Redirect(url);
	}

	/**
	 * Creates {@code JSONResponse} to send the specified entity to the client 
	 * as JSON format response.
	 * 
	 * @param value The object to be sent to the client.
	 * @return {@code JSONResponse}.
	 */
	public static JSONResponse json(Object value) {
		return new JSONResponse(new Result(value));
	}
	
	/**
	 * Creates {@code XMLResponse} to send the specified entity to the client as 
	 * XML format response.
	 * 
	 * @param value The object to be sent to the client.
	 * @return {@code XMLResponse}.
	 */
	public static XMLResponse xml(Object value) {
		return new XMLResponse(new Result(value));
	}
	
}
