/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony;

/**
 * {@code Settings} represents Monophony framework's settings.
 * You can specify your custom settings by declaring {@code ServletContext} 
 * init parameter as follow in web.xml.
 * <pre>
 * &lt;context-param&gt;
 * 		&lt;param-name&gt;org.eiichiro.monophony.settings&lt;/param-name&gt;
 * 		&lt;param-value&gt;com.example.your.app.CustomSettings;/param-value&gt;
 * &lt;/context-param&gt;
 * </pre>
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public interface Settings {

	/**
	 * Returns {@code Request} implementation class to represent the specified 
	 * content type. The {@code Request} implementation class must have the 
	 * constructor that accept <code>WebContext</code> as parameter.
	 * 
	 * @param contentType The content type specified in HTTP request header. 
	 * This parameter may be <code>null</code>.
	 * @return The {@code Request} implementation class corresponding to the 
	 * specified content type.
	 */
	public Class<? extends Request> requestType(String contentType);
	
	/**
	 * Returns {@code Response} implementation class to represent the specified 
	 * content type. The {@code Response} implementation class must have the 
	 * constructor that accept <code>Result</code> as parameter. 
	 * 
	 * @param contentType The content type specified in HTTP response header. 
	 * This parameter may be <code>null</code>.
	 * @return The {@code Response} implementation class corresponding to the 
	 * specified content type.
	 */
	public Class<? extends Response> responseType(String contentType);
	
	/**
	 * Returns {@code Loader} instance to load Web endpoint classes.
	 * 
	 * @return The loader instance to load Web endpoint classes.
	 */
	public Loader loader();
	
	/**
	 * Returns {@code Instantiator} instance to instantiate Web endpoint 
	 * instance.
	 * 
	 * @return The instantiator instance to instantiate Web endpoint instance.
	 */
	public Instantiator instantiator();
	
}
