/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony.xml;

import java.lang.reflect.Field;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Collection;

import org.eiichiro.monophony.AbstractParameter;
import org.eiichiro.monophony.Parameter;
import org.eiichiro.monophony.Types;
import org.eiichiro.monophony.WebContext;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * {@code XMLBodyParameter} abstracts Web endpoint method parameter in HTTP 
 * request body source as XML format.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class XMLBodyParameter extends AbstractParameter {

	private final Node node;
	
	/**
	 * Constructs a new {@code XMLBodyParameter} with the specified Web context, 
	 * parameter type, parameter name and DOM document object.
	 * 
	 * @param context Web context.
	 * @param type The parameter type.
	 * @param name The parameter name.
	 * @param node The parsed DOM node.
	 */
	public XMLBodyParameter(WebContext context, Type type, String name, Node node) {
		super(context, type, name, false);
		this.node = node;
		construct();
	}
	
	/**
	 * Constructs the value of Web endpoint parameter from HTTP request body 
	 * sent as XML format.
	 * {@code XMLBodyParameter} supports the following parameter declaration 
	 * independently. Other than them are the same as {@code AbstractParameter}.
	 * <ol>
	 * <li>Named collection of user-defined object type</li>
	 * <li>Named user-defined object type</li>
	 * </ol>
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void construct() {
		if (name != null && !name.isEmpty()) {
			if (Types.isCollection(type)) {
				if (Types.isSupportedCollection(type)) {
					Class<?> elementType = Types.getElementType(type);
					
					if (!Types.isCoreValueType(elementType)
							&& !Types.isUserDefinedValueType(elementType)) {
						// Named collection of user-defined object type.
						Collection<Object> nodes = getValues(name);
						
						if (nodes == null) {
							logger.debug("Collection named [" + name + "] not found");
							return;
						}
						
						try {
							Class<?> implementationType = Types.getDefaultImplementationType(type);
							Collection<Object> collection = (Collection<Object>) implementationType.newInstance();
							
							try {
								for (Object node : nodes) {
									Object instance = elementType.newInstance();
									
									for (Field field : elementType.getDeclaredFields()) {
										Object object = new XMLBodyParameter(
												context, field.getGenericType(),
												field.getName(),
												(Node) node).getValue();
										
										if (object != null) {
											field.setAccessible(true);
											field.set(instance, object);
										}
									}
									
									collection.add(instance);
								}
								
								value = (collection.isEmpty()) ? null : collection;
							} catch (Exception e) {
								logger.debug("Cannot instantiate ["
										+ Types.getDefaultImplementationType(type)
										+ "] (Default implementation type of ["
										+ type + "])", e);
							}
							
						} catch (Exception e) {
							logger.warn("Cannot instantiate [" + elementType
									+ "] (Collection element type of [" + type
									+ "])", e);
						}
						
						return;
					}
				}
				
			} else if (!Types.isCoreValueType(type)
					&& !Types.isUserDefinedValueType(type)) {
				// Named user-defined object type.
				Class<?> rawType = Types.getRawType(type);
				Object node = getValue(name);
				
				try {
					Object instance = rawType.newInstance();
					
					for (Field field : rawType.getDeclaredFields()) {
						Object object = new XMLBodyParameter(context, field
								.getGenericType(), field.getName(),
								(Node) node).getValue();
						
						if (object != null) {
							field.setAccessible(true);
							field.set(instance, object);
						}
					}
					
					value = instance;
				} catch (Exception e) {
					logger.warn("Cannot instantiate [" + type + "]", e);
				}
				
				return;
			}
		}
		
		super.construct();
	}
	
	/**
	 * Converts the specified object to the specified type.
	 * The object to be converted must be a {@code Node} that has no child, or 
	 * this method return <code>null</code>.
	 * 
	 * @param object The object to be converted
	 * @param type The type to which the specified object be converted.
	 * @return The converted object.
	 */
	@Override
	protected Object convert(Object object, Class<?> type) {
		if (object instanceof Node) {
			Node node = (Node) object;
			NodeList nodes = node.getChildNodes();
			Text text = null;
			
			for (int i = 0; i < nodes.getLength(); i++) {
				Node child = nodes.item(i);
				
				if (child.getNodeType() == Node.TEXT_NODE) {
					text = (Text) child;
				}
			}
			
			if (text == null) {
				logger.warn("Parameter [" + object + "] cannot be converted to ["
						+ type + "]; Converted 'object' must have one child ["
						+ Text.class + "] node");
				return null;
			}
			
			return super.convert(text.getNodeValue(), type);
		} else {
			logger.warn("Parameter [" + object + "] cannot be converted to ["
					+ type + "]; Converted 'object' must be a ["
					+ Node.class + "]");
			return null;
		}
	}
	
	/**
	 * Converts the specified object to the specified user-defined value type.
	 * The object to be converted must be a {@code Node} that has no child, or 
	 * this method return <code>null</code>.
	 * 
	 * @param object The object to be converted
	 * @param type The user-defined value type to which the specified object be 
	 * converted.
	 * @return The converted object.
	 */
	@Override
	protected Object convertUserDefinedValueType(Object object, Class<?> type) {
		if (object instanceof Node) {
			Node node = (Node) object;
			NodeList nodes = node.getChildNodes();
			Text text = null;
			
			for (int i = 0; i < nodes.getLength(); i++) {
				Node child = nodes.item(i);
				
				if (child.getNodeType() == Node.TEXT_NODE) {
					text = (Text) child;
				}
			}
			
			if (text == null) {
				logger.warn("Parameter [" + object + "] cannot be converted to ["
						+ type + "]; Converted 'object' must have one child ["
						+ Text.class + "] node");
				return null;
			}
			
			return super.convertUserDefinedValueType(text.getNodeValue(), type);
		} else {
			logger.warn("Parameter [" + object + "] cannot be converted to ["
					+ type + "]; Converted 'object' must be a ["
					+ Node.class + "]");
			return null;
		}
	}
	
	/**
	 * Returns the Web endpoint method parameter value corresponding to the 
	 * specified name from XML content in HTTP request body.
	 * 
	 * @param name The parameter name.
	 * @return The Web endpoint method parameter value corresponding to the 
	 * specified name as {@code Node}.
	 * @see org.eiichiro.monophony.AbstractParameter#getValue(java.lang.String)
	 */
	@Override
	protected Object getValue(String name) {
		NodeList nodes = node.getChildNodes();
		
		for (int i = 0; i < nodes.getLength(); i++) {
			Node item = nodes.item(i);
			
			if (item.getNodeType() == Node.ELEMENT_NODE
					&& item.getNodeName().equals(name)) {
				return item;
			}
		}
		
		return null;
	}

	/**
	 * Returns the Web endpoint method parameter value corresponding to the 
	 * specified name as a collection from XML content in HTTP request body.
	 * 
	 * @param name The parameter name.
	 * @return The Web endpoint method parameter value corresponding to the 
	 * specified name as collection of {@code Node}.
	 * @see org.eiichiro.monophony.AbstractParameter#getValues(java.lang.String)
	 */
	@Override
	protected Collection<Object> getValues(String name) {
		Collection<Object> collection = new ArrayList<Object>();
		NodeList nodes = node.getChildNodes();
		
		for (int i = 0; i < nodes.getLength(); i++) {
			Node item = nodes.item(i);
			
			if (item.getNodeType() == Node.ELEMENT_NODE
					&& item.getNodeName().equals(name)) {
				collection.add(item);
			}
		}
		
		return (collection.isEmpty()) ? null : collection;
	}

	/**
	 * Returns a new {@code XMLBodyParameter} with the specified Web context, 
	 * parameter type and parameter name.
	 * 
	 * @param context Web context.
	 * @param type The parameter type.
	 * @param name The parameter name.
	 * 
	 * @see org.eiichiro.monophony.AbstractParameter#newParameter(org.eiichiro.monophony.WebContext, java.lang.reflect.Type, java.lang.String)
	 */
	@Override
	protected Parameter newParameter(WebContext context, Type type, String name) {
		return new XMLBodyParameter(context, type, name, node);
	}

}
