/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.monophony.xml;

import java.lang.reflect.Type;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.eiichiro.monophony.GenericRequest;
import org.eiichiro.monophony.Parameter;
import org.eiichiro.monophony.WebContext;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

/**
 * {@code XMLRequest} is a XML-based implementation of {@code Request}.
 * If the requested MIME media type is 'application/xml', default settings of 
 * Monophony uses this implementation. In this request type, Web endpoint method 
 * parameter in {@code @Body} source is represented by {@code XMLBodyParameter}.
 * 
 * @see XMLBodyParameter
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class XMLRequest extends GenericRequest {

	private final Node node;
	
	/**
	 * Constructs a new {@code XMLRequest} with the specified Web context.
	 * 
	 * @param context The Web context.
	 */
	public XMLRequest(WebContext context) {
		super(context);
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setNamespaceAware(true);
		Node node = null;
		
		try {
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document document = builder.parse(new InputSource(context.getRequest().getReader()));
			node = document.getDocumentElement();
		} catch (Exception e) {
			logger.warn("Cannot parse XML document into DOM object", e);
		}
		
		this.node = node;
	}

	/**
	 * Returns a new {@code XMLBodyParameter} with the specified Web context, 
	 * parameter type, parameter name and DOM document object in this request.
	 * 
	 * @param context Web context.
	 * @param type The parameter type.
	 * @param name The parameter name.
	 */
	@Override
	protected Parameter newBodyParameter(WebContext context, Type type, String name) {
		return new XMLBodyParameter(context, type, name, node);
	}
	
}
